#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <math.h>
#include <time.h>

#ifdef _WIN32
#pragma warning( disable : 4127 ) /* disable conditional is constant warning */
#endif

#define t 1  // true
#define f 0  //false
#define version "4.02 - October 29th, 2005"

#define MAX 256000
#define MAX1 66000

void writeresults(mpz_t p, mpz_t d) // write the intermediate results
{
FILE *out; //file results.txt

	out=fopen("results.txt","a+");
	fprintf(out,"M");
	gmp_fprintf(out,"%Zd",p);
	fprintf(out," has a factor: ");
	gmp_fprintf(out,"%Zd",d);
        gmp_fprintf(out,"\n");
	fclose(out);
}

void writeendfactor (mpz_t p, float start, float stop) // write the final results
{
FILE *out; //file results.txt

	out=fopen("results.txt","a+");
	fprintf(out,"M");
	gmp_fprintf(out,"%Zd",p);
	fprintf(out," no factor from %3.3f bits to %3.3f bits.",start,stop);
        fprintf(out,"\n");
	printf("\n");
        printf("M");
	gmp_printf("%Zd",p);
	printf(" no factor from %3.3f bits to %3.3f bits.",start,stop);
        printf("\n");
	fclose(out);
}


// write parameters in status.txt

void writestatus(mpz_t p, mpz_t k, mpz_t start_bit, mpz_t stop_bit, float start, float stop)
{
FILE *file; // file status.txt

	file = fopen ("status.txt", "w");

	gmp_fprintf(file,"%Zd",p); 		// save p
        gmp_fprintf(file,"\n");
	gmp_fprintf(file,"%Zd",k); 		// save k
        gmp_fprintf(file,"\n");
	gmp_fprintf(file,"%Zd",start_bit);	// save start_bit
        gmp_fprintf(file,"\n");
	gmp_fprintf(file,"%Zd",stop_bit); 	// save stop_bit
        gmp_fprintf(file,"\n");
	gmp_fprintf(file,"%3.3f",start);	// save bds
        gmp_fprintf(file,"\n");
	gmp_fprintf(file,"%3.3f",stop);		// save bde
        gmp_fprintf(file,"\n");

	fclose(file);
}

void usage () {
        printf("\n");
	gmp_printf("usage : ./factor4 [-parameter]");
        printf("\n");
        printf("\n");
	gmp_printf("   -h : this screen");
        printf("\n");
	gmp_printf("   -f : get value and bit_depth of first factor");
        printf("\n");
	gmp_printf("   -b : run benchmark");
        printf("\n");
	gmp_printf("   -r : resume status.txt file in silent mode");
        printf("\n");
        printf("\n");
	
	exit(0);
}

void factor_depth() {
    mpz_t exponent, bit_depth,d,k;
    
    mpz_init_set_ui(exponent,1);
    mpz_init_set_ui(bit_depth,1);
    mpz_init_set_ui(d,1);
    mpz_init_set_ui(k,1);

    gmp_printf("\nFactor4 v.%s ",version);
    printf("\n");
    printf("Please enter the exponent to be tested: ");
    gmp_scanf("%Zd",exponent);    
    printf("Please enter the k value to be tested: ");
    gmp_scanf("%Zd",k);       
    if (mpz_probab_prime_p(exponent,30)==0) mpz_nextprime(exponent, exponent);
    mpz_set(d,exponent);
    mpz_mul_ui(d,exponent,2);
    mpz_mul(d, d,k);
    mpz_add_ui(d, d, 1);
    printf("\n");
    gmp_printf ("For k = %Zd and exponent = %Zd --> d = 2kp+1 = %Zd",k,exponent,d); 
    printf("\n");
    NF_writebits(d);

}



int bench() {
    int iterazioni = 100000;
    int i,j,fact0  = 0;
    
    mpz_t d1[20], base, d_b, k_b, prodotto_b, exponent_b, tw_b;
    
    double begin,end;
    
    gmp_printf("\nFactor4 v.%s ",version);
    printf("\n");
    gmp_printf("Timing %i iterations at different bit length: ",iterazioni);
    // Init variables
    mpz_init_set_ui(d_b,0);
    mpz_init_set_ui(k_b,1);
    mpz_init_set_ui(prodotto_b,1);
    mpz_init_set_ui(base,2);
    mpz_init2(tw_b,34);
    mpz_init_set_ui(exponent_b,3321928097);  
    
    //???
    mpz_mul_ui(tw_b,exponent_b,2);
    // Index exponents
    mpz_init_set_str (d1[ 0],"205959542014000001",10)  ; // 57 bits
    mpz_init_set_str (d1[ 1],"411919084028000001",10)  ; // 58 bits
    mpz_init_set_str (d1[ 2],"830482024250000001",10)  ; // 59 bits
    mpz_init_set_str (d1[ 3],"1660964048500000001",10)  ; // 60 bits
    mpz_init_set_str (d1[ 4],"3321928097000000001",10)  ; // 61 bits
    mpz_init_set_str (d1[ 5],"6643856194000000001",10)  ; // 62 bits
    mpz_init_set_str (d1[ 6],"13287712388000000001",10)  ; // 63 bits
    mpz_init_set_str (d1[ 7],"26575424776000000001",10)  ; // 64 bits
    mpz_init_set_str (d1[ 8],"53150849552000000001",10)  ; // 65 bits
    mpz_init_set_str (d1[ 9],"106301699104000000001",10)  ; // 66 bits
    mpz_init_set_str (d1[10],"212603398208000000001",10)  ; // 67 bits
    mpz_init_set_str (d1[11],"425206796416000000001",10)  ; // 68 bits
    mpz_init_set_str (d1[12],"850413592832000000001",10)  ; // 69 bits
    mpz_init_set_str (d1[13],"1660964048500000000001",10)  ; // 70 bits
    mpz_init_set_str (d1[14],"3321928097000000000001",10)  ; // 71 bits
    mpz_init_set_str (d1[15],"6643856194000000000001",10)  ; // 72 bits
    mpz_init_set_str (d1[16],"13287712388000000000001",10)  ; // 73 bits
    mpz_init_set_str (d1[17],"26575424776000000000001",10)  ; // 74 bits
    mpz_init_set_str (d1[18],"53150849744000000000001",10)  ; // 75 bits
    mpz_init_set_str (d1[19],"106301699104000000000001",10)  ; // 76 bits
    for (j=0; j<20;j++) {
        begin = clock();
        mpz_set(d_b,d1[j]);
        mpz_mul_ui(k_b,k_b,20);
        i=0;
        while (i<iterazioni) {
               mpz_add_ui(k_b,k_b,i);
               mpz_mul(d_b,k_b,tw_b);
               mpz_add_ui(d_b,d_b,1); 
               mpz_powm(prodotto_b,base,exponent_b,d1[j]);
               if ((mpz_cmp_ui(prodotto_b,1))==0) { 
                   fact0 = 1;
               } 
            i++;
        }    
        end = clock();
        end = (((end-begin)/CLOCKS_PER_SEC) + ((end-begin)/CLOCKS_PER_SEC)*0.31429);
        printf("\n");
        gmp_printf("Done %i iterations at %i bits: %3.2f sec",iterazioni,j+57,end); 
    }
    mpz_clear(d_b);
    mpz_clear(k_b);
    mpz_clear(prodotto_b);
    mpz_clear(base);
    mpz_clear(tw_b);
}

int MersennePrime(mpz_t x)
{
int		    mers[42]={1,3,5,7,13,17,19,31,61,89,107,127,521,607,1279,2203,2281,
                      3217,4253,4423,9689,9941,11213,19937,21701,23209,44497,
                      86243,110503,132049,216091,756839,859433,1257787,1398269,
                      2976221,3021377,6972593,13466917,20996011,24036583,25964951}; // Known Mersenne Primes
int i=0, j=0, k=0;

   k=mpz_get_ui(x);
   for (i=0;i<42;i++) {
      if (mers[i]==k) j=1;
   };
   return j;
}

/*
void writebits(mpz_t d) {

int     i,nbit = 0;
char    s[200];
mpz_t   resto,uno1,c;

    mpz_init_set_ui(uno1,1);    
    mpz_init_set_ui(resto,0);    
    mpz_init_set_ui(c,0);
    mpz_get_str(s,2,d);
    nbit=strlen(s)-1; 
    mpz_mul_2exp(c,uno1,nbit); // c = 2^(nbit-1)
    mpz_sub_ui(c,c,1); // c = 2^(nbit-1) - 1
    mpz_set(resto,d); // reminder = d
    mpz_sub(resto,resto,c); // reminder = reminder-c = d2 - c
    mpz_mul_ui(resto,resto,1000); // reminder = reminder*1000 
    mpz_tdiv_q(resto,resto,c); // reminder = (reminder*1000)/2^(nbit-1)-1
    i = mpz_get_ui(resto);
    if (i>0) i=(int)(log10(i)); 
      else i=0;
    printf("\r\t\t\t\t\t\t\t    %i.",nbit);
    switch (i) {
      case (0): printf("00"); break;
      case (1): printf("0"); break;
    }
    gmp_printf("%Zd",resto); printf(" bit depth");
    mpz_clear(uno1);    
    mpz_clear(resto);    
    mpz_clear(c);
}
*/

int NF_writebits(mpz_t d) {
int nbit = 0;
double d2,bits;

    nbit=(mpz_sizeinbase(d,2)-1);
    if (nbit<1000) {
        d2=mpz_get_d(d);
        bits=log2(d2);
        printf("\r");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("    %.5g bit depth  ",bits);
    } else {
        printf("\r");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("\t");
        printf("    %i bit depth   ",nbit);
    }
}



int main (int argc, char **argv) 
{
 
 mpz_t        exponent,
              twiceexponent,
              start_bit,
              stop_bit,
              inizio,
              fine,
              uno,
              g,
              k,k1,
              d,d1,
              b,
              prodotto, risultato;
              
 mpz_t        kmin, kmax, kn, iter;  // iteration flags
              
 mpz_t        check[MAX1], // MAX1 precalculated 2kp+1 values for MOD operation
              bitmap,
              base;              
 
 mpz_t        risultato1,pivot;
              
 int          a,indice;
 int          folder[MAX],   // 256000 values for sieve
              sieve[MAX/10],  // 25600  prime values 
              FlagP[MAX/10];  // 25600 flags for primes
 int          modulo, 
              modulo1, 
              MAXPRIMES;      
              
 int          last_iter, last, step; // iteration flags      
 int          BIT_DEPTH = 1024000000;
              
 // Modulo 120: eliminates 3 o 5 mod 8 and multiples of 3 and 5 
 int sequenza [16][60] =   {{3,5,7,9,11,13,15,17,19,21,23,25,27,29,31,33,35,37,39,41,43,45,47,49,51,53,55,57,59,61,63,65,67,69,71,73,75,77,79,81,83,85,87,89,91,93,95,97,99,101,103,105,107,109,111,113,115,117,119,1},
                            {15,29,43,57,71,85,99,113,7,21,35,49,63,77,91,105,119,13,27,41,55,69,83,97,111,5,19,33,47,61,75,89,103,117,11,25,39,53,67,81,95,109,3,17,31,45,59,73,87,101,115,9,23,37,51,65,79,93,107,1},
                            {23,45,67,89,111,13,35,57,79,101,3,25,47,69,91,113,15,37,59,81,103,5,27,49,71,93,115,17,39,61,83,105,7,29,51,73,95,117,19,41,63,85,107,9,31,53,75,97,119,21,43,65,87,109,11,33,55,77,99,1},
                            {27,53,79,105,11,37,63,89,115,21,47,73,99,5,31,57,83,109,15,41,67,93,119,25,51,77,103,9,35,61,87,113,19,45,71,97,3,29,55,81,107,13,39,65,91,117,23,49,75,101,7,33,59,85,111,17,43,69,95,1},
                            {35,69,103,17,51,85,119,33,67,101,15,49,83,117,31,65,99,13,47,81,115,29,63,97,11,45,79,113,27,61,95,9,43,77,111,25,59,93,7,41,75,109,23,57,91,5,39,73,107,21,55,89,3,37,71,105,19,53,87,1},
                            {39,77,115,33,71,109,27,65,103,21,59,97,15,53,91,9,47,85,3,41,79,117,35,73,111,29,67,105,23,61,99,17,55,93,11,49,87,5,43,81,119,37,75,113,31,69,107,25,63,101,19,57,95,13,51,89,7,45,83,1},
                            {47,93,19,65,111,37,83,9,55,101,27,73,119,45,91,17,63,109,35,81,7,53,99,25,71,117,43,89,15,61,107,33,79,5,51,97,23,69,115,41,87,13,59,105,31,77,3,49,95,21,67,113,39,85,11,57,103,29,75,1},
                            {59,117,55,113,51,109,47,105,43,101,39,97,35,93,31,89,27,85,23,81,19,77,15,73,11,69,7,65,3,61,119,57,115,53,111,49,107,45,103,41,99,37,95,33,91,29,87,25,83,21,79,17,75,13,71,9,67,5,63,1},
                            {63,5,67,9,71,13,75,17,79,21,83,25,87,29,91,33,95,37,99,41,103,45,107,49,111,53,115,57,119,61,3,65,7,69,11,73,15,77,19,81,23,85,27,89,31,93,35,97,39,101,43,105,47,109,51,113,55,117,59,1},
                            {75,29,103,57,11,85,39,113,67,21,95,49,3,77,31,105,59,13,87,41,115,69,23,97,51,5,79,33,107,61,15,89,43,117,71,25,99,53,7,81,35,109,63,17,91,45,119,73,27,101,55,9,83,37,111,65,19,93,47,1},
                            {83,45,7,89,51,13,95,57,19,101,63,25,107,69,31,113,75,37,119,81,43,5,87,49,11,93,55,17,99,61,23,105,67,29,111,73,35,117,79,41,3,85,47,9,91,53,15,97,59,21,103,65,27,109,71,33,115,77,39,1},
                            {87,53,19,105,71,37,3,89,55,21,107,73,39,5,91,57,23,109,75,41,7,93,59,25,111,77,43,9,95,61,27,113,79,45,11,97,63,29,115,81,47,13,99,65,31,117,83,49,15,101,67,33,119,85,51,17,103,69,35,1},
                            {95,69,43,17,111,85,59,33,7,101,75,49,23,117,91,65,39,13,107,81,55,29,3,97,71,45,19,113,87,61,35,9,103,77,51,25,119,93,67,41,15,109,83,57,31,5,99,73,47,21,115,89,63,37,11,105,79,53,27,1},
                            {99,77,55,33,11,109,87,65,43,21,119,97,75,53,31,9,109,85,63,41,19,117,95,73,51,29,7,105,83,61,39,17,115,93,71,49,27,5,103,81,59,37,15,113,91,69,47,35,3,101,79,57,35,13,111,89,67,45,23,1},
                            {107,93,79,65,51,37,23,9,115,101,87,73,59,45,31,17,3,109,95,81,67,53,39,25,11,117,103,89,75,61,47,33,19,5,111,97,83,69,55,41,27,13,119,105,91,77,63,49,35,21,7,113,99,85,71,57,43,29,15,1},
                            {119,117,115,113,111,109,107,105,103,101,99,97,95,93,91,89,87,85,83,81,79,77,75,73,71,69,67,65,63,61,59,57,55,53,51,49,47,45,43,41,39,37,35,33,31,29,27,25,23,21,19,17,15,13,11,9,7,5,3,1},
                           };
                         
 int     i,j; 
 int     fact0 = 0;         // flag to factors in the number

 time_t begin;             
 double end;

 double temp;
 int bds1=2, bde1=1, bde2 = 1;
 float bds=2.0, bde=1.0;
 int w;


 FILE	*file;

      if (argc > 2) usage();
      if (argc == 2) { 
	      if (argv[1][0] != '-') usage();
      }
      mpz_init_set_ui(exponent,0);
      mpz_init_set_ui(twiceexponent,0);      
      mpz_init_set_ui(start_bit,0);
      mpz_init_set_ui(stop_bit,0);
      mpz_init_set_ui(inizio,0);
      mpz_init_set_ui(fine,0);
      mpz_init_set_ui(uno,1);
      mpz_init_set_ui(g,0);
      mpz_init_set_ui(k,0);
      mpz_init_set_ui(k1,1); 
      mpz_init_set_ui(d,0);
      mpz_init_set_ui(d1,1);
      mpz_init_set_ui(b,2);
      mpz_init_set_ui(prodotto,1);
      mpz_init_set_ui(risultato,1);
      
      mpz_init_set_ui(bitmap,0);
      mpz_init_set_ui(base,2);
      mpz_init_set_ui(iter,0);
      
      mpz_init_set_ui(risultato1,1);
      mpz_init_set_ui(pivot,0);

      for (w=0;w<MAX1;w++) mpz_init_set_ui(check[w],1);
      
       
    if (argc == 1) { 
	    gmp_printf("\nFactor4 v. %s",version);
            printf("\n");
	    printf("(Mersenne numbers factorization written by Luigi Morelli)"); 
            printf("\n");        
	    gmp_printf("Please enter the exponent to be factored: 2^");
	    gmp_scanf("%Zd",exponent);
	    mpz_mul_ui(twiceexponent,exponent,2);	    
	    if (mpz_cmp_ui(exponent,0)==0) {goto end1;};
	    mpz_set_ui(g,79999999);
	    if (mpz_cmp(g,exponent)==1) {
        	if (MersennePrime(exponent)==1) {
		       	printf("You entered a known Mersenne Prime number. Try Again!\n"); 
			goto end1;
		};
	    }
	    if ((mpz_probab_prime_p(exponent,20)==0)){ 
        	printf("You didn't enter a prime number. Try Again!\n");
	        goto end1;
	    } 
	    
	    // test_how_deep_factoring;

	    gmp_printf("Now enter start bit depth  : "); gmp_scanf(" %f",&bds);
	    gmp_printf("Finally enter end bit depth : "); gmp_scanf(" %f",&bde); 
	    if (bds<2.0) bds=2.0; 
            printf("\n");
	    gmp_printf("Sieving factors from %3.3f bits up to %3.3f bits...",bds,bde);
            printf("\n");
	    begin = time(NULL);              //*****new line*****//

	    // Definition of start_bit, stop_bit
        // mpz_mul_2exp (start_bit, uno, bds); 
	    // mpz_mul_2exp (stop_bit, uno, bde); 
	    bds1 = trunc(bds);   
	    bds = bds - bds1;
	    bds = bds * 100000000;                         // la parte decimale viene moltiplicata per 1000000
	    bde2 = floor(bds);                            // viene eliminata la parte decimale  
	    mpz_set_ui(risultato, bde2);                 // la parte decimale, moltiplicata per 1M e arrotondata a intero in risultato
	    mpz_mul_2exp (start_bit, uno, bds1);         // start_bit con la parte intera di bds
	    mpz_mul_2exp (risultato1, risultato, bds1);  // risultato 1 = risultato * 2^bds
	    mpz_fdiv_q_ui(risultato1, risultato1, 100000000);// viene rinormalizzato risultato1
	    mpz_add(start_bit, start_bit, risultato1);   // start_bit = 2^(int)bit + parte decimale bit * 2^(int)bit
	    
	    bde1 = trunc(bde);
	    bde = bde - bde1;
	    bde = bde * 100000000;
	    bde2 = ceil(bde);
	    mpz_set_ui(risultato, bde2);
	    mpz_mul_2exp (stop_bit, uno, bde1);         
	    mpz_mul_2exp (risultato1, risultato, bde1);  
	    mpz_fdiv_q_ui(risultato1, risultato1, 100000000);
	    mpz_add(stop_bit, stop_bit, risultato1); 
	    
	    bds = bds / 100000000; bds = bds + bds1;
	    bde = bde / 100000000; bde = bde + bde1;
 	    
	    mpz_set(g,exponent);
	    mpz_mul_ui(g,g,2);
	    mpz_tdiv_q(g,g,start_bit);
	    if (mpz_cmp_ui(start_bit,1)!=0) 
    	    mpz_sub_ui(start_bit, start_bit,1);
	    mpz_set(d,start_bit);
	    mpz_sub_ui(d,d,1);
	    mpz_tdiv_q_ui(d,d,2);
	    mpz_tdiv_q(k,d,exponent); // Define starting k from start_bit
    } else {
        switch (argv[1][1]) {
            case 'f': {
                factor_depth();
                goto end1;
            }
            case 'b': {
                bench();
                goto end1;
            }
            case 'r': {
                    gmp_printf("Factor4 v. %s",version);
                    printf("\n");
         	    printf("(Mersenne numbers factorization written by Luigi Morelli)");         
          	    file = fopen("status.txt","r");			
          	    gmp_fscanf(file,"%Zd",exponent);	// load exponent
          	    gmp_fscanf(file,"%Zd",k);		    // load actual k
          	    gmp_fscanf(file,"%Zd",start_bit);	// load actual start_bit
          	    gmp_fscanf(file,"%Zd",stop_bit);	// load actual stop_bit
          	    fscanf(file,"%f",&bds);		        // load bds
          	    fscanf(file,"%f",&bde);		        // load bde
          	    fclose(file);
          	    mpz_mul_ui(twiceexponent,exponent,2);
                    printf("\n");
          	    gmp_printf("Resuming factoring of %Zd",exponent);
          	    printf(" up to %3.3f bit depth...",bde);
                    printf("\n");
         	    break;
       	    }
       	    default: {
       	        usage();
       	        break;
   	        }
   	    }
    }
    
    // bds, bde contain the initial and final bit depth as power of 2 (2^bds, 2^bde)
    // start_bit and stop_bit contain the range of the potential factors to check
    // We use those values to set the number of needed iterations of k
    // kn holds the number of k's, iter the number of BIT_DEPTH iterations, 
    // last_iter the number of last round (kn mod BIT_DEPTH)
    
    mpz_init_set(kmin, k);
    mpz_init_set(kmax, stop_bit);
    mpz_sub_ui(kmax, kmax, 1);
    mpz_div_ui(kmax, kmax, 2);
    mpz_div(kmax, kmax, exponent);
    mpz_init_set_ui(kn,1);
    mpz_sub(kn, kmax, kmin);
    
    mpz_fdiv_q_ui(iter, kn, BIT_DEPTH);      // if range < BIT_DEPTH, iter = 0
    last_iter = mpz_fdiv_r_ui(risultato, kn, BIT_DEPTH);
    
    // Starting 2kp+1 from chosen bit depth //
    if (mpz_cmp_ui(k,0)==0) mpz_set_ui(k,1); 
    mpz_set(d,k);     
    mpz_mul_ui(d,d,2); 
    mpz_mul(d,d,exponent); 
    mpz_add_ui(d,d,1);     
    //**************************************//


    // Eratosthenes sieve : sieve[0] = 2; MAXPRIMES = j
    indice=2;
    j=0;
    for (i=0;i<MAX;i++) {folder[i] = t;}
    while (indice < MAX) { 
        a = indice-2;
        if (folder[a] == t) {
            sieve[j] = indice;
            FlagP[j] = t;
            j++;
            while (a < MAX-indice) {
                a=a+indice;
                folder[a] = f;
            } 
        }
        indice++;
    }
    // MAXPRIMES = j;
    MAXPRIMES = ((bde+8.79)*100000000)/1000000;
    if (MAXPRIMES > j) MAXPRIMES = j; 
    
    // Start loop
        last=0;
start1:
    while (mpz_cmp(pivot,iter)<0) {
    
       for (i=0;i<MAXPRIMES+1;i++) {FlagP[i] = t;}

       temp=clock();

       // Create an array with the first 66.000 test factors of the form 2kp+1 starting from actual k,d
       printf("\n");
       printf("Reinitializing sieve");
       printf("\n");
       mpz_set(d1,d);
       mpz_set(check[0],d1);     // check[0] holds 2kp+1 with k=1
       for (a=1;a<MAX1;a++) {
	       mpz_add(d1,d1,twiceexponent);
           mpz_set(check[a],d1);
       }
    

       // Select the bitmap for the sieve - The bitmap (1000000...0000 has 0 for prime and 1 for composite
       mpz_pow_ui(bitmap,base,BIT_DEPTH);
    


    // Select the 1 or 7 modulo 8 values and checks if k=1
    // mpz_init_set(k1,k);
    // mpz_mul(k1,exponent,k1);
    // mpz_add_ui(k1,k1,1);             
    
       // checks factor modulo 8 for the first 60 factors
       modulo  = mpz_fdiv_ui(check[0],8); 
           switch (modulo) {
               case 1:
                   if (mpz_fdiv_ui(check[1],8) == 3) {
                       i=1;
                   } else { // modulo = 7
                       i=2;
                   }
               break;
               case 3:
                   if (mpz_fdiv_ui(check[1],8) == 5) { 
                       i=0;
                   } else { // modulo 1
                       mpz_setbit(bitmap,0);
                       i=3;
                   }               
               break;
               case 5:
                   if (mpz_fdiv_ui(check[1],8) == 7) {  
                       mpz_setbit(bitmap,0);
                       i=3;
                   } else { // modulo 3
                       i=0;
                   }          
               break;
               case 7:
                   if (mpz_fdiv_ui(check[1],8) == 1) {
                       i=2;
                   } else { // modulo 5
                       i=1;
                   }
               break;
           }  
       while (i<60) {
           mpz_setbit (bitmap,i);
           i++;
           mpz_setbit (bitmap,i);
           i++;
           i++;
           i++;     
       };
        
       // Look for factor mod 120 = 1: i is the index
       i = 0;       
       modulo = mpz_fdiv_r_ui(risultato,check[0],120);
       while (modulo != 1) {
          i++;
          modulo = mpz_fdiv_r_ui(risultato,check[i],120); 
       } 
    
       // Now i holds position of 2kp+1 mod 120 = 1, with k = i+1
       modulo1 = mpz_fdiv_r_ui(risultato,check[i+1],120);  
    
       // sequenza [i][j]: look for i position (sequence type [0-15])
       w = 0;
       while (modulo1 != sequenza[w][0]) w++; 
   
       // The index w shows the correct sequence and the bit-setting start algorithm for sequenza[w][i] 
       // (i = k-1 = first value of 2kp+1 with module = 1 mod 120)
       j = BIT_DEPTH-60;
       switch (w) {   
        case 0:    
            while (i<j) {
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++; mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);  
                i++;
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               
            }         
            break;    
        case 1: 
            while (i<j) {
                i++;mpz_setbit(bitmap,i);  
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             
            }
            break;
        case 2:    
            while (i<j) {
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;            
            }
            break; 
        case 3: 
            while (i<j) {
                i++; mpz_setbit(bitmap,i); // 27
                i++; mpz_setbit(bitmap,i); // 53
                i++;
                i++; mpz_setbit(bitmap,i); // 105
                i++; mpz_setbit(bitmap,i); // 11
                i++; mpz_setbit(bitmap,i); // 37
                i++; mpz_setbit(bitmap,i); // 63
                i++;
                i++; mpz_setbit(bitmap,i); // 115
                i++; mpz_setbit(bitmap,i); // 21
                i++;
                i++;
                i++; mpz_setbit(bitmap,i); // 99
                i++; mpz_setbit(bitmap,i); // 5
                i++;
                i++; mpz_setbit(bitmap,i); // 57
                i++; mpz_setbit(bitmap,i); // 83
                i++; mpz_setbit(bitmap,i); // 109
                i++; mpz_setbit(bitmap,i); // 15
                i++;
                i++; mpz_setbit(bitmap,i); // 67
                i++; mpz_setbit(bitmap,i); // 93
                i++;
                i++; mpz_setbit(bitmap,i); // 25
                i++; mpz_setbit(bitmap,i); // 51   
                i++; mpz_setbit(bitmap,i); // 77
                i++;
                i++; mpz_setbit(bitmap,i); // 9
                i++; mpz_setbit(bitmap,i); // 35
                i++; mpz_setbit(bitmap,i); // 61
                i++; mpz_setbit(bitmap,i); // 87
                i++;
                i++; mpz_setbit(bitmap,i); // 19
                i++; mpz_setbit(bitmap,i); // 45
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);                                                                                                                                                
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);                                
                i++;
                i++; mpz_setbit(bitmap,i); // 33
                i++; mpz_setbit(bitmap,i); // 59
                i++; mpz_setbit(bitmap,i); // 85
                i++; mpz_setbit(bitmap,i); // 111
                i++;
                i++; mpz_setbit(bitmap,i); // 43
                i++; mpz_setbit(bitmap,i); // 69
                i++; mpz_setbit(bitmap,i); // 95
                i++;                                                                                                                                                                                                             
            }        
            break;
        case 4: 
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }        
            break;            
        case 5: 
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }        
            break;
        case 6:
            while (i<j) {
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               
            }         
            break;
        case 7:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }         
            break; 
        case 8:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);                
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                
            }         
            break;
        case 9:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);                
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_tstbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }         
            break;
        case 10:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    
            }         
            break;
        case 11: 
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
            }        
            break; 
        case 12:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }         
            break;
        case 13:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
            }         
            break;
        case 14:
            while (i<j) {
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         
            }         
            break;
        case 15:
            while (i<j) {
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;
                i++; mpz_setbit(bitmap,i);
                i++; mpz_setbit(bitmap,i);
                i++;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    
            }         
            break;                                                                       
       }    
       /*************** Fine Modulo 120 ******************/
  
       j=3;
       while (j<MAXPRIMES) {
           i=0;
           do {
               a = mpz_fdiv_ui(check[i], sieve[j]);
               i++;
           } while ((a!=0) && (i<MAX1));
           if (i<MAX1) {
               i--; 
               indice = i;
               w = BIT_DEPTH - sieve[j];
               while (indice < w) {
                   indice = indice + sieve[j];
                   if (mpz_tstbit(bitmap,indice) == 0)
                       mpz_setbit(bitmap,indice);
               }
           }
           j++;
       }


       end = clock();
       printf("\n");
       printf("Sieve time   : %3.3f secs. ",(float)((end-temp)/CLOCKS_PER_SEC));
       i=0;
       for (w=0;w<BIT_DEPTH;w++) if (mpz_tstbit(bitmap,w)==1) i++;
       temp = ((float)i*100)/BIT_DEPTH;
       gmp_printf ("sieve percentage : %3.3f",temp); 
       printf("\n");

       // Start check of remaining possible factors 2kp+1 

       i=0;
       step = BIT_DEPTH / 20;
       while (i<BIT_DEPTH) {
           for (w=0;w<step;w++) {  
               if (i>BIT_DEPTH) break;   
               // Extraction from the bitmap
               if (mpz_tstbit(bitmap,i) == 0) { 
                   // The value represents a possible factor of 2^p-1
                   mpz_add_ui(k1,k,i);
                   mpz_mul(d,k1,twiceexponent);
                   mpz_add_ui(d,d,1); 
                   mpz_powm(prodotto,base,exponent,d);
                   if ((mpz_cmp_ui(prodotto,1))==0) { 
                       printf("\r");
                       printf("M"); gmp_printf("%Zd",exponent); printf(" has a factor: "); gmp_printf("%Zd",d);                                      NF_writebits(d); printf("\n");
                       writeresults(exponent,d);
                       fact0 = 1;
                   } 
               }
               i++;         
           }
           printf("\n"); 
           gmp_printf("k=%Zd, d=%Zd",k1,d); NF_writebits(d); 
       }

       // control loop for writing status on disk
       writestatus(exponent,k1,start_bit,stop_bit,bds,bde);        
       mpz_set(k,k1);
       
       end = clock();
       temp = end-temp;
       
       mpz_add_ui(pivot, pivot, 1);
                
    } // end main loop
    
    if (last==0) {
        BIT_DEPTH = last_iter;
        mpz_add_ui(iter, iter,1);
        last = 1; 
        goto start1;
    }
    
    if (fact0 == 0) writeendfactor(exponent,bds,bde);    

    end = clock();
    printf("\n");
    gmp_printf("Total time   : %3.3f secs.",(float)(end/CLOCKS_PER_SEC));         //*****new line*****//
    printf("\n");

      mpz_clear(exponent);
      mpz_clear(twiceexponent);      
      mpz_clear(uno);
      mpz_clear(start_bit);
      mpz_clear(stop_bit);
      mpz_clear(inizio);
      mpz_clear(fine);
      mpz_clear(g);
      mpz_clear(k);
      mpz_clear(d);
      mpz_clear(b);
      mpz_clear(prodotto);
      mpz_clear(risultato);
      
      mpz_clear(bitmap);
      mpz_clear(base);
      mpz_clear(iter);
      
      mpz_clear(risultato1);
      mpz_clear(pivot);

    return 0;
    end1:
    return 1;
}
